#!/usr/bin/env python
#-*-mode: python; encoding: utf-8-*-

#=========================================================================
"""
  The ``dimgx`` Python Module
  ===========================

  Utilities for programmatically inspecting `Docker
  <https://www.docker.com/whatisdocker/>`_ `images
  <https://docs.docker.com/terms/image/>`__ and extracting their `layers
  <https://docs.docker.com/terms/layer/>`__.

  Notice
  ------

  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the ``LICENSE`` (or ``LICENSE.txt``) file which accompanied
  this software for rights and restrictions governing its use. All rights
  not expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.

  Examples
  --------

  Extract all but the first layer of image ``0fe1d2c3b4a5`` to a file
  called ``output.tar`` in the current working directory (without proper
  error checking)::

        import docker
        import os
        import tarfile
        from dimgx import *

        DEFAULT_DOCKER_HOST = docker.utils.utils.DEFAULT_UNIX_SOCKET
        DOCKER_HOST = os.environ.get('DOCKER_HOST', DEFAULT_DOCKER_HOST)
        dc = docker.Client(DOCKER_HOST)

        IMAGE_ID = '0fe1d2c3b4a5' # short version
        layers_dict = inspectlayers(dc, IMAGE_ID)
        with tarfile.open('output.tar', 'w') as tar_file:
          extractlayers(dc, layers_dict[':layers'][1:], tar_file)

  Extract all but the first and last layers *in reverse order* (note the
  ``0`` provided for the :obj:`top_most_layer` parameter to
  :func:`extractlayers`)::

      with tarfile.open('output.tar', 'w') as tar_file:
          layers_to_extract = layers_dict[':layers'][1:-1]
          layers_to_extract.reverse()
          extractlayers(dc, layers_to_extract, tar_file, top_most_layer=0)

  Module Contents
  ---------------
  """
#=========================================================================

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

from builtins import * # pylint: disable=redefined-builtin,unused-wildcard-import,wildcard-import
from future.builtins.disabled import * # pylint: disable=redefined-builtin,unused-wildcard-import,wildcard-import

#---- Imports ------------------------------------------------------------

from copy import deepcopy
from datetime import datetime
from io import open
from logging import (
    ERROR,
    getLogger,
)
from posixpath import (
    basename as posixpath_basename,
    commonprefix as posixpath_commonprefix,
    dirname as posixpath_dirname,
    join as posixpath_join,
    sep as posixpath_sep,
)
from os.path import (
    join as path_join,
    realpath as path_realpath,
)
from shutil import rmtree
from tarfile import open as tarfile_open
from tempfile import mkdtemp
from dateutil.parser import parse as dateutil_parse
from humanize import naturalsize
from _dimgx import (
    TZ_UTC,
    UnsafeTarPath,
    logexception,
    naturaltime,
)
from _dimgx.version import __version__

#---- Constants ----------------------------------------------------------

__all__ = (
    'UnsafeTarPath',
    'denormalizeimage',
    'extractlayers',
    'inspectlayers',
    'normalizeimage',
)

_LOGGER = getLogger(__name__)
_WHITEOUT_PFX = '.wh.'
_WHITEOUT_PFX_LEN = len(_WHITEOUT_PFX)

#---- Functions ----------------------------------------------------------

#=========================================================================
def denormalizeimage(image_desc, copy=False):
    """
    :param image_desc: a normalized image description as returned from
                       |docker.Client.images|_,
                       |docker.Client.inspect_image|_, etc., and passed to
                       :func:normalizeimage

    :param copy: if :const:`True`, make a copy of :obj:`image_desc` before
                 performing any denormalizations

    Removes any entries created by :func:normalizeimage.
    """
    if copy:
        image = deepcopy(image_desc)
    else:
        image = image_desc

    remove = [ k for k in image if k.startswith(':') ]

    for k in remove:
        del image[k]

    return image

#=========================================================================
def extractlayers(dc, layers, tar_file, top_most_layer=-1):
    """
    :param dc: a |docker.Client|_

    :param layers: a sequence of inspection objects (likely retrieved with
                   :func:`inspectlayers`) corresponding to the layers to
                   extract and flatten

    :param tar_file: a :class:`~tarfile.TarFile` open for writing to which
                     to write the flattened layer archive

    :param top_most_layer: an image ID or an index into :obj:`layers`
                           indicating the most recent layer to retrieve
                           (the default of ``-1`` references the last item
                           in :obj:`layers`; see below)

    :raises: :class:`docker.errors.APIError` or
             :class:`docker.errors.DockerException` - on failure
             interacting with Docker (e.g., bad image ID, failed
             connection, Docker not running, etc.); or

             :class:`UnsafeTarPath` - probably indicative of a bug in
             Docker

    Retrieves the layers corresponding to the :obj:`layers` parameter and
    extracts them into :obj:`tar_file`. Changes from layers corresponding
    to larger indexes in :obj:`layers` will overwrite or block those from
    smaller ones.

    Callers will need to set the :obj:`top_most_layer` parameter if
    :obj:`layers` is not in ascending order. It is always safe to provide
    the same value as the :obj:`image_spec` parameter to
    :func:`inspectlayers`, but this may be ineffecient if that layer does
    not appear in :obj:`layers`.

    .. |docker.Client| replace:: :class:`docker.Client`
    .. _`docker.Client`: https://docker-py.readthedocs.org/en/latest/api/
    """
    if not layers:
        _LOGGER.warning('nothing to extract')

        return

    image_spec = top_most_layer if not isinstance(top_most_layer, int) else layers[top_most_layer][':id']
    tmp_dir = path_realpath(mkdtemp())

    try:
        image = logexception(_LOGGER, ERROR, 'unable to retrieve image layers from "{}": {{e}}'.format(image_spec), dc.get_image, image_spec)

        with tarfile_open(mode='r|*', fileobj=image) as image_tar_file:
            next_info = image_tar_file.next()

            while next_info:
                next_path = path_realpath(path_join(tmp_dir, next_info.name))

                if not next_path.startswith(tmp_dir):
                    exc = UnsafeTarPath('unsafe path: "{}"'.format(next_info.name))
                    logexception(_LOGGER, ERROR, 'unable to retrieve entry from export of "{}": {{e}}'.format(image_spec), exc)

                image_tar_file.extract(next_info, tmp_dir)
                next_info = image_tar_file.next()

        seen = set()
        hides_subtrees = set()

        # Look through each layer's archive (oldest to newest)
        for layer in layers[::-1]:
            layer_id = layer[':id']
            layer_tar_path = path_join(tmp_dir, layer_id, 'layer.tar')

            with tarfile_open(layer_tar_path) as layer_tar_file:
                next_info = layer_tar_file.next()

                while next_info:
                    next_dirname = posixpath_dirname(next_info.name)
                    next_basename = posixpath_basename(next_info.name)

                    if next_basename.startswith(_WHITEOUT_PFX):
                        removed_path = posixpath_join(next_dirname, next_basename[_WHITEOUT_PFX_LEN:])
                        hides_subtrees.add(( removed_path, 'removal' ))

                        if removed_path in seen:
                            _LOGGER.debug('skipping removal "%s"', removed_path)
                        else:
                            _LOGGER.debug('hiding "%s" as removed', removed_path)
                    elif next_info.name in seen:
                        _LOGGER.debug('skipping "%s" as overwritten', next_info.name)
                    else:
                        next_name_len = len(next_info.name)
                        hidden = None

                        for h, deverbal in hides_subtrees: # https://en.wikipedia.org/wiki/deverbal
                            if len(h) > next_name_len:
                                continue

                            common_pfx = posixpath_commonprefix(( h, next_info.name ))
                            common_pfx_len = len(common_pfx)

                            if next_name_len == common_pfx_len \
                                    or next_info.name[common_pfx_len:].startswith(posixpath_sep):
                                hidden = deverbal, h
                                break

                        if hidden:
                            _LOGGER.debug('skipping "%s" hidden by %s of %s', next_info.name, *hidden)
                        else:
                            mtime = naturaltime(datetime.utcfromtimestamp(next_info.mtime).replace(tzinfo=TZ_UTC))
                            _LOGGER.info('writing "%s" from "%s" to archive (size: %s; mode: %o; mtime: %s)', next_info.name, layer_id, naturalsize(next_info.size), next_info.mode, mtime)

                            if next_info.linkname:
                                # TarFile.extractfile() tries to do
                                # something weird when its parameter
                                # represents a link (see the docs)
                                fileobj = None
                            else:
                                fileobj = layer_tar_file.extractfile(next_info)

                            tar_file.addfile(next_info, fileobj)
                            seen.add(next_info.name)

                            if not next_info.isdir():
                                hides_subtrees.add(( next_info.name, 'presence' ))

                    next_info = layer_tar_file.next()
    finally:
        rmtree(tmp_dir, ignore_errors=True)

#=========================================================================
def inspectlayers(dc, image_spec):
    """
    :param dc: a |docker.Client|_

    :param image_spec: the name or ID of the image to inspect

    :returns: a :class:`dict` containing the descriptions (see below)

    :raises: :class:`docker.errors.APIError` or
             :class:`docker.errors.DockerException` on failure
             interacting with Docker

    Retrieves and normalizes descriptions for the :obj:`image_spec` image
    and each of its ancestors by calling |docker.Client.images|_.

    The returned :class:`dict` is as follows::

        {
            ':layers': ( image_desc_0, ... image_desc_n ),
            image_id_0: 0,
            ...
            image_id_n: n,
            ...
        }

    The :attr:`':layers'` :class:`tuple` is in ascending order (i.e., from
    the root to :obj:`image_spec`). The other entries map the layers'
    various IDs, to their respective indexes in the :attr:`':layers'`
    :class:`tuple`.

    .. |docker.Client| replace:: :class:`docker.Client`
    .. _`docker.Client`: https://docker-py.readthedocs.org/en/latest/api/
    .. |docker.Client.images| replace:: :func:`docker.Client.images`
    .. _`docker.Client.images`: https://docker-py.readthedocs.org/en/latest/api/#images
    """
    images = logexception(_LOGGER, ERROR, 'unable to retrieve image summaries: {{e}}'.format(), dc.images, all=True)
    images_dict = {}

    for image in images:
        normalizeimage(image)
        image_id = image[':id']
        image_short_id = image[':short_id']
        images_dict[image_id] = image
        images_dict[image_short_id] = image

        for repo_tag in image[':repo_tags']:
            images_dict[repo_tag] = image

    if image_spec not in images_dict:
        raise RuntimeError('{} not found among the layers retreieved for that image'.format(image_spec))

    layers = []
    layer = images_dict[image_spec]
    layer_id = layer[':id']

    if image_spec not in ( layer_id, layer[':short_id'] ):
        _LOGGER.debug('image "%s" has ID "%s"', image_spec, layer[':short_id'])

    while True:
        layers.append(layer)
        parent_layer_short_id = layer[':parent_id'][:12]

        if not parent_layer_short_id:
            _LOGGER.debug('found root layer "%s"', layer[':short_id'])
            break

        layer_id = parent_layer_short_id
        layer = images_dict[layer_id]

    layers.reverse()
    layers_by_id = {
        ':all_images': images_dict,
        ':layers': tuple(layers),
    }

    for i, layer in enumerate(layers):
        layers_by_id[layer[':id']] = i
        layers_by_id[layer[':short_id']] = i

        for repo_tag in layer[':repo_tags']:
            layers_by_id[repo_tag] = i

    return layers_by_id

#=========================================================================
def normalizeimage(image_desc, copy=False):
    """
    :param image_desc: an image description as returned from
                       |docker.Client.images|_,
                       |docker.Client.inspect_image|_, etc.

    :param copy: if :const:`True`, make a copy of :obj:`image_desc` before
                 performing any normalizations

    :returns: the normalized image description (:obj:`image_desc` if
              :obj:`copy` is :const:`False`)

    This method is attempts to address certain `Docker API inconsistencies
    <https://github.com/docker/docker/issues/5893#issuecomment-102398746>`__.
    The following keys are added to :obj:`image_desc`:

    * :attr:`':id'` - a normalized :attr:`'Id'`
    * :attr:`':short_id'` - the first 12 hexidecimal characters from
      :attr:`':id'`
    * :attr:`':parent_id'` - a normalized :attr:`'ParentId'` or
      :attr:`'Parent'`
    * :attr:`':created_dt'` - a timezone-aware :class:`datetime` object
      representing :attr:`'Created'`
    * :attr:`':repo_tags'` - a normalized :attr:`'RepoTags'`, including
      any short names (i.e., those implying ``:latest``)

    .. |docker.Client.images| replace:: :py:func:`docker.Client.images`
    .. _`docker.Client.images`: https://docker-py.readthedocs.org/en/latest/api/#images
    .. |docker.Client.inspect_image| replace:: :py:func:`docker.Client.inspect_image`
    .. _`docker.Client.inspect_image`: https://docker-py.readthedocs.org/en/latest/api/#inspect_image
    """
    if copy:
        image = deepcopy(image_desc)
    else:
        image = image_desc

    image_id = image.get('Id', image.get('id')).lower()
    image[':id'] = image_id
    image[':parent_id'] = image.get('ParentId', image.get('Parent', image.get('parent', ''))).lower()
    image_short_id = image_id[:12]
    image[':short_id'] = image_short_id
    image_created = image.get('Created', image.get('created'))

    if isinstance(image_created, int):
        # Work-around for
        # <https://github.com/PythonCharmers/python-future/issues/144> and
        # <https://bitbucket.org/pypy/pypy/issue/2048/datetimeutcfromtimestamp-barfs-when>
        if hasattr(image_created, '__native__'):
            image_created = image_created.__native__()

        image[':created_dt'] = datetime.utcfromtimestamp(image_created).replace(tzinfo=TZ_UTC)
    else:
        image[':created_dt'] = dateutil_parse(image_created)

    repo_tags = image.get('RepoTags', ())
    image[':repo_tags'] = [ t for t in repo_tags if t != '<none>:<none>' ]

    for repo_tag in repo_tags:
        repo, tag = repo_tag.split(':')

        if tag == 'latest':
            image[':repo_tags'].append(repo)

    return image

#---- Initialization -----------------------------------------------------

if __name__ == '__main__':
    from _dimgx.cmd import main as _main
    _main()



#-*-mode: python; encoding: utf-8-*-

#=========================================================================
"""
  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the ``LICENSE`` (or ``LICENSE.txt``) file which accompanied
  this software for rights and restrictions governing its use. All rights
  not expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.
"""
#=========================================================================

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

from builtins import * # pylint: disable=redefined-builtin,unused-wildcard-import,wildcard-import
from future.builtins.disabled import * # pylint: disable=redefined-builtin,unused-wildcard-import,wildcard-import

#---- Imports ------------------------------------------------------------

from datetime import datetime
from logging import ERROR
from os import linesep
from traceback import (
    format_exception,
    format_stack,
)
from dateutil.tz import tzutc
from humanize import naturaltime as _naturaltime
from requests import RequestException

#---- Constants ----------------------------------------------------------

__all__ = ()

TZ_UTC = tzutc()

#---- Exceptions ---------------------------------------------------------

#=========================================================================
class UnsafeTarPath(Exception):
    """
    Raised before an entry is extracted from a Docker tar archive if it
    has an unsafe path (usually either an absolute path, or one that
    contains "``..``" or equivalent).
    """

#---- Functions ----------------------------------------------------------

#=========================================================================
logexception = lambda *args, **kw: _logexception(*args, **kw) # pylint: disable=unnecessary-lambda

#=========================================================================
def naturaltime(value, future=False, months=True):
    if isinstance(value, datetime):
        if value.tzinfo is None:
            now = datetime.now()
        else:
            now = datetime.utcnow().replace(tzinfo=TZ_UTC)

        value = now - value

    return _naturaltime(value, future, months)

#=========================================================================
def _debugtrace(logger, lvl, msg, last_type, last_value, last_tb):
    logger.log(lvl, msg)
    stack = format_stack()[:-2] + format_exception(last_type, last_value, last_tb)[1:-1]
    val = 'Complete traceback (most recent call last):' + linesep \
        + ''.join(( str(l) for l in stack ))
    logger.debug(val.rstrip())

#=========================================================================
def _logexception(logger, lvl, msg_fmt, f, *args, **kw):
    try:
        if isinstance(f, Exception):
            raise f

        return f(*args, **kw)
    except EnvironmentError as e:
        if e.strerror:
            msg_e = e.strerror
        else:
            msg_e = e

        last_type, last_value, last_tb = sys.exc_info()

        try:
            _debugtrace(logger, lvl, msg_fmt.format(e=msg_e), last_type, last_value, last_tb)
        finally:
            del last_type, last_value, last_tb # break circular references

        if lvl >= ERROR:
            raise
    except Exception as e: # pylint: disable=broad-except
        last_type, last_value, last_tb = sys.exc_info()

        try:
            _debugtrace(logger, lvl, msg_fmt.format(e=e), last_type, last_value, last_tb)
        finally:
            del last_type, last_value, last_tb # break circular references

        if lvl >= ERROR:
            raise



#-*-mode: python; encoding: utf-8-*-

#=========================================================================
"""
  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the ``LICENSE`` (or ``LICENSE.txt``) file which accompanied
  this software for rights and restrictions governing its use. All rights
  not expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.
"""
#=========================================================================

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

#---- Constants ----------------------------------------------------------

__all__ = ()

__version__ = '0.1.4'



#-*-mode: python; encoding: utf-8-*-

#=========================================================================
"""
  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the ``LICENSE`` (or ``LICENSE.txt``) file which accompanied
  this software for rights and restrictions governing its use. All rights
  not expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.
"""
#=========================================================================

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

from builtins import * # pylint: disable=redefined-builtin,unused-wildcard-import,wildcard-import
from future.builtins.disabled import * # pylint: disable=redefined-builtin,unused-wildcard-import,wildcard-import

#---- Imports ------------------------------------------------------------

from argparse import (
    ArgumentParser,
    ArgumentTypeError,
)
from collections import OrderedDict
from errno import errorcode
from functools import wraps
from logging import (
    DEBUG,
    ERROR,
    INFO,
    WARNING,
    basicConfig as logging_basicConfig,
    getLevelName as logging_getLevelName,
    getLogger,
)
from os import (
    O_CREAT,
    O_EXCL,
    O_TRUNC,
    O_WRONLY,
    SEEK_CUR,
    environ,
    fdopen,
    fstat,
    lseek,
    open as os_open,
)
from os.path import (
    extsep as ospath_extsep,
    splitext as ospath_splitext,
)
from re import (
    IGNORECASE,
    compile as re_compile,
)
from stat import S_ISREG
from sys import (
    argv as sys_argv,
    exit as sys_exit,
    stdout,
)
from tarfile import open as tarfile_open
from docker import Client
from docker.utils.utils import DEFAULT_UNIX_SOCKET as DOCKER_DEFAULT_UNIX_SOCKET
from humanize import naturalsize
from dimgx import (
    inspectlayers,
    extractlayers as dimgx_extractlayers,
)
from _dimgx import (
    logexception,
    naturaltime,
)

#---- Constants ----------------------------------------------------------

__all__ = ()

_LOGGER = getLogger(__name__)
_LAYER_RE_STR = r'(?:0|-?[1-9]\d{0,10}|-?0[Bb][01]+|-?0[Oo][0-7]+|-?0[Xx][0-9A-Fa-f]+|[0-9A-Fa-f]{12,64})'
_LAYER_SPEC_RE = re_compile(r'^(?P<l>{layer_re})(?::(?P<r>{layer_re}))?$'.format(layer_re=_LAYER_RE_STR), IGNORECASE)
_TARGET_STDOUT = '-'
_CMP_BZIP2 = 'bz2'
_CMP_GZIP = 'gz'
_CMP_NONE = None
_DEFAULT_CMP_LVL = 9
_DEFAULT_LOG_FMT = '%(levelname)-8s: %(message)s'
_LOG_LEVELS_BY_NAME = OrderedDict(( ( logging_getLevelName(l), l ) for l in ( ERROR, WARNING, INFO, DEBUG ) ))
_DEFAULT_LOG_LVL = logging_getLevelName(WARNING)
_EXIT_EXEC = 2
_EXIT_LAYER_SPEC = 3

_USAGE = """
%(prog)s [options] [-l LAYER_SPEC] ... [-t PATH] IMAGE_SPEC
%(prog)s -h # for help
"""

_DESCRIPTION = """
Extract and flatten Docker image layers.
"""

_EPILOG = """
Diagnostics: 1 for a problem encountered while parsing the command arguments; {} for an execution error; {} for a LAYER_SPEC that doesn\'t reference a known layer (if strict layer specifications are enabled).
""".format(_EXIT_EXEC, _EXIT_LAYER_SPEC)

_LAYER_GROUP_DESCRIPTION = """
Layers can be specified as indexes [-n..n) (where "n" is the total number of layers present in the image), or Docker image IDs.
Negative indexes count from the end (i.e., "-i" is equivalent to "n - i").
Layers may also be specified as ranges in the form "i:j" (where "i" and "j" are indexes or image IDs).
The range "j:i" is equivalent to the reverse of the range "i:j".
Ranges are inclusive, meaning "i:i" is equivalent to "i".
If no layers are specified, all layers are inferred, starting with the root (i.e., primogenitor), and ending with the specified image (i.e., "0:-1").
Layers are extracted in order, which means that files from layers specified later in the command line will overwrite (or block) those from earlier ones where there is a conflict.
Ordering is resolved before retrieval so that each distinct layer is only extracted once.
"""

_TARGET_GROUP_DESCRIPTION = """
If no target is provided, information about the specified layers is written to STDOUT, one line per layer.
If a target is provided, the specified layers will be extracted and written to the target as a tar archive.
"""

#---- Decorators ---------------------------------------------------------

#=========================================================================
def exitonraise(f):
    @wraps(f)
    def _wrapper(*_args, **_kw):
        try:
            return f(*_args, **_kw)
        except: # pylint: disable=bare-except
            sys_exit(_EXIT_EXEC)

    return _wrapper

#---- Functions ----------------------------------------------------------

#=========================================================================
def buildparser():
    parser = ArgumentParser(description=_DESCRIPTION, epilog=_EPILOG, usage=_USAGE)
    parser.add_argument('image', help='the name or ID of the Docker image', metavar='IMAGE_SPEC')

    layer_group = parser.add_argument_group(description=_LAYER_GROUP_DESCRIPTION)
    layer_group.add_argument('-l', '--layers', action='append', help='the selected layer(s) (defaults to all layers in ascending order)', metavar='LAYER_SPEC', type=layertype)
    layer_group.add_argument('-r', '--reverse', action='store_true', dest='reverse', help='reverse the layer order from that specified')
    layer_group.add_argument('-R', '--no-reverse', action='store_false', dest='reverse', help='preserve the layer order as specified (default)')
    layer_group.add_argument('-s', '--strict', action='store_true', dest='strict', help='treat any layer specification that doesn\'t reference at least one known layer as an error')
    layer_group.add_argument('-S', '--no-strict', action='store_false', dest='strict', help='ignore any layer specification that doesn\'t reference a known layer (default))')

    target_group = parser.add_argument_group(description=_TARGET_GROUP_DESCRIPTION)
    target_group.add_argument('-t', '--target', action='store', help='the path to which to write the archive ("{}" for STDOUT)'.format(_TARGET_STDOUT.replace('%', '%%')), metavar='PATH')
    target_group.add_argument('-q', '--quiet', action='store_true', dest='quiet', help='when no target is specified, print only the image IDs')
    target_group.add_argument('-Q', '--no-quiet', action='store_false', dest='quiet', help='when no target is specified, print the image IDs with additional information in a table (default)')
    target_group.add_argument('-w', '--force', action='store_true', dest='force', help='overwrite the target archive if it already exists')
    target_group.add_argument('-W', '--no-force', action='store_false', dest='force', help='don\'t overwrite the target archive if it already exists (default)')

    environ_docker_host = environ.get('DOCKER_HOST')
    docker_host = environ_docker_host if environ_docker_host else DOCKER_DEFAULT_UNIX_SOCKET
    docker_group = parser.add_argument_group()
    docker_group.add_argument('-d', '--docker-host', default=docker_host, help='the Docker host URI (defaults to "{}" or DOCKER_HOST, if set)'.format(DOCKER_DEFAULT_UNIX_SOCKET.replace('%', '%%')), metavar='URI')

    compress_group = parser.add_argument_group()
    compress_group.add_argument('-j', '--bzip2', action='store_const', const=_CMP_BZIP2, dest='compression', help='compress the target archive with bzip2 compression')
    compress_group.add_argument('-z', '--gzip', action='store_const', const=_CMP_GZIP, dest='compression', help='compress the target archive with gzip compression')
    compress_group.add_argument('-C', '--no-compress', action='store_const', const=_CMP_NONE, default=_CMP_NONE, dest='compression', help='do not compress the target archive (default)')
    compress_group.add_argument('--compress-level', choices=list(range(10)), default=_DEFAULT_CMP_LVL, help='the desired compression level for the archive (defaults to {})'.format(_DEFAULT_CMP_LVL), metavar='0..9', type=int)

    log_group = parser.add_argument_group()
    log_group.add_argument('--log-level', choices=list(_LOG_LEVELS_BY_NAME), default=_DEFAULT_LOG_LVL, help='the desired logging level (defaults to "{}")'.format(_DEFAULT_LOG_LVL.replace('%', '%%')))
    log_group.add_argument('--log-format', default=_DEFAULT_LOG_FMT, help='a logging format compatible with Python\'s "logging" module (defaults to "{}")'.format(_DEFAULT_LOG_FMT.replace('%', '%%')), metavar='FORMAT_SPEC')

    return parser

#=========================================================================
def extractlayers(dc, args, layers, top_most_layer_id):
    target_path = args.target
    flags = O_WRONLY

    if target_path == _TARGET_STDOUT:
        target_fd = stdout.fileno()
    else:
        flags |= O_CREAT | O_TRUNC

        if not args.force:
            flags |= O_EXCL

        target_fd = logexception(_LOGGER, ERROR, 'unable to open target file "{}": {{e}}'.format(target_path), os_open, target_path, flags, 0o666)

    with fdopen(target_fd, 'wb') as target_file:
        if hasattr(target_file, 'seekable'):
            seekable = target_file.seekable()
        else:
            try:
                seekable = not lseek(target_fd, 0, SEEK_CUR) < 0 \
                    and S_ISREG(fstat(target_fd).st_mode)
            except OSError as e:
                if errorcode.get(e.errno) != 'ESPIPE':
                    raise

                seekable = False

        open_args = { 'fileobj': target_file }

        if args.compression is None:
            open_args['mode'] = 'w' if seekable else 'w|'
        else:
            if seekable:
                mode = 'w:{}'
                open_args['compresslevel'] = args.compress_level
                _, ext = ospath_splitext(target_path)

                if ext.lower() != '{}{}'.format(ospath_extsep, args.compression):
                    _LOGGER.warning('target name "%s" doesn\'t match compression type ("%s")', target_path, args.compression)
            else:
                mode = 'w|{}'
                _LOGGER.warning('target "%s" is not seekable, ignoring compression level (%d)', target_path, args.compress_level)

            open_args['mode'] = mode.format(args.compression)

        with tarfile_open(**open_args) as tar_file:
            dimgx_extractlayers(dc, layers, tar_file, top_most_layer_id)

#=========================================================================
def layerspec2index(args, layers, layer_spec_part):
    if layer_spec_part is None:
        return None

    num_layers = len(layers[':layers'])

    if len(layer_spec_part) < 12:
        try:
            li = int(layer_spec_part, 0)

            if not args.strict:
                new_li = min(max(li, -num_layers), num_layers - 1)

                if new_li != li:
                    li = new_li
                    _LOGGER.warning('"%s" is out of range (truncating) to "%d")', layer_spec_part, li)

            if li >= -num_layers \
                    and li < num_layers:
                return li if li >= 0 else num_layers + li
        except ValueError:
            pass
    else:
        try:
            return layers[layer_spec_part]
        except KeyError:
            pass

    no_layer_msg = '"%s" does not resolve to any layer associated with image "%s"'

    if args.strict:
        _LOGGER.error(no_layer_msg, layer_spec_part, args.image)
        sys_exit(_EXIT_LAYER_SPEC)

    _LOGGER.warning(no_layer_msg, layer_spec_part, args.image)

    return None

#=========================================================================
def layertype(value):
    matches = _LAYER_SPEC_RE.search(value)

    if not matches:
        raise ArgumentTypeError('"{}" is not a valid LAYER_SPEC'.format(value))

    return ( matches.group('l'), matches.group('r'), value )

#=========================================================================
def main():
    import _dimgx
    _dimgx._logexception = exitonraise(_dimgx._logexception) # WARNING: monkey patch; pylint: disable=protected-access
    args = buildparser().parse_args(sys_argv[1:])
    logging_basicConfig(format=args.log_format)
    getLogger().setLevel(logging_getLevelName(args.log_level))
    dc = Client(args.docker_host)
    layers_dict = inspectlayers(dc, args.image)
    top_most_layer_id, selected_layers = selectlayers(args, layers_dict)

    if not selected_layers:
        _LOGGER.warning('no known layers selected')

    if args.target is None:
        printlayerinfo(args, selected_layers)
    else:
        extractlayers(dc, args, selected_layers, top_most_layer_id)

#=========================================================================
def printlayerinfo(args, layers, outfile=stdout):
    if args.quiet:
        for l in layers:
            print(l['Id'][:12].lower(), file=outfile)

        return

    total_size = 0
    fields_fmt = '\t'.join([ '{:<15}' ] * 5)
    print(fields_fmt.format('IMAGE ID', 'PARENT ID', 'CREATED', 'LAYER SIZE', 'VIRTUAL SIZE'), file=outfile)

    for l in layers:
        image_id = l['Id'][:12].lower()
        parent_id = l[':parent_id'][:12].lower()

        if not parent_id:
            parent_id = '-'

        created = naturaltime(l[':created_dt'])
        layer_size = naturalsize(l['Size'])
        total_size += l['Size']
        virt_size = naturalsize(total_size)
        print(fields_fmt.format(image_id, parent_id, created, layer_size, virt_size), file=outfile)

#=========================================================================
def selectlayers(args, layers):
    layer_specs = args.layers

    if layer_specs is None:
        selected_indexes = list(range(len(layers[':layers'])))
    else:
        selected_indexes = []
        last_num_selected_indexes = num_selected_indexes = len(selected_indexes)

        for l, r, v in layer_specs:
            last_num_selected_indexes = num_selected_indexes
            li = layerspec2index(args, layers, l)
            ri = layerspec2index(args, layers, r)

            if li is None:
                continue

            if r is None:
                selected_indexes.append(li)
            elif ri is None:
                continue
            elif ri < li:
                selected_indexes.extend(reversed(range(ri, li + 1))) # upper bounds are inclusive
            else:
                selected_indexes.extend(range(li, ri + 1)) # upper bounds are inclusive

            num_selected_indexes = len(selected_indexes)

            if num_selected_indexes == last_num_selected_indexes:
                empty_layer_range_msg = '"%s" resolves to an empty range'

                if args.strict:
                    _LOGGER.error(empty_layer_range_msg, v)
                    sys_exit(_EXIT_LAYER_SPEC)
                else:
                    _LOGGER.warning(empty_layer_range_msg, v)

    if args.reverse:
        selected_indexes.reverse()

    # Take the last of each index specified (so we don't have to look at
    # each distinct layer more than once)
    seen = OrderedDict()

    for i in selected_indexes[::-1]:
        if i not in seen:
            seen[i] = None # use OrderedDict as an ordered set

    top_most_layer_id = None if not seen else layers[':layers'][max(seen)][':id']
    selected_layers = [ layers[':layers'][i] for i in seen ][::-1]

    return top_most_layer_id, selected_layers



#!/usr/bin/env python
#-*-mode: python; encoding: utf-8-*-

#=========================================================================
"""
  Copyright |(c)| 2015 `Matt Bogosian`_ (|@posita|_).

  .. |(c)| unicode:: u+a9
  .. _`Matt Bogosian`: mailto:mtb19@columbia.edu
  .. |@posita| replace:: **@posita**
  .. _`@posita`: https://github.com/posita

  Please see the ``LICENSE`` (or ``LICENSE.txt``) file which accompanied
  this software for rights and restrictions governing its use. All rights
  not expressly waived or licensed are reserved. If such a file did not
  accompany this software, then please contact the author before viewing
  or using this software in any capacity.
"""
#=========================================================================

from __future__ import (
    absolute_import, division, print_function,
    # See <http://bugs.python.org/setuptools/issue152>
    # unicode_literals,
)

#---- Imports ------------------------------------------------------------

from setuptools import (
    find_packages,
    setup,
)

# See this e-mail thread:
# <http://www.eby-sarna.com/pipermail/peak/2010-May/003348.html>
import logging # pylint: disable=unused-import
import multiprocessing # pylint: disable=unused-import

from inspect import (
    currentframe,
    getframeinfo,
)
from os.path import (
    dirname,
    isfile,
    join as ospath_join,
)

#---- Constants ----------------------------------------------------------

__all__ = ()

INSTALL_REQUIRES = (
    'docker-py',
    'future',
    'humanize',
    'python-dateutil',
)

_MY_DIR = dirname(getframeinfo(currentframe()).filename)

#---- Initialization -----------------------------------------------------

_namespace = {
    '__version__': '<none>',
    '_version_path': ospath_join(_MY_DIR, '_dimgx', 'version.py'),
}

if isfile(_namespace['_version_path']):
    exec(compile(open(_namespace['_version_path']).read(), _namespace['_version_path'], 'exec'), _namespace, _namespace) # pylint: disable=exec-used

_SETUP_ARGS = {
    'name'                : 'dimgx',
    'version'             : _namespace['__version__'],
    'author'              : 'Matt Bogosian',
    'author_email'        : 'mtb19@columbia.edu',
    'url'                 : 'https://github.com/posita/py-dimgx',
    'bugtrack_url'        : 'https://github.com/posita/py-dimgx/issues',
    'license'             : 'MIT License',
    'description'         : 'extract and flatten Docker image layers',
    'long_description'    : open(ospath_join(_MY_DIR, 'README.rst')).read(),

    # From <http://pypi.python.org/pypi?%3Aaction=list_classifiers>
    'classifiers': (
        'Topic :: Software Development :: Build Tools',
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: System :: Archiving :: Packaging',
    ),

    'packages'            : find_packages(exclude = ( 'tests', )),
    'py_modules'          : ( 'dimgx', ),
    'include_package_data': True,
    'install_requires'    : INSTALL_REQUIRES,
    'test_suite'          : 'tests',

    'entry_points': {
        'console_scripts': (
            'dimgx = _dimgx.cmd:main',
        ),
    },
}

if __name__ == '__main__':
    setup(**_SETUP_ARGS)
